from coralillo import Model, Form, BoundedModel, fields
from coralillo.validation import validation_rule
from coralillo.errors import InvalidFieldError
import sys
import inspect


# for model testing
class Table(Model):
    name = fields.Text()


class Ship(Model):
    name = fields.Text()
    code = fields.Text(index=True)


class Tenanted(BoundedModel):
    name = fields.Text()

    @classmethod
    def prefix(cls):
        return 'testing'


class SideWalk(Model):
    name = fields.Text()


class House(Model):
    number = fields.Integer(required=False)


# for events test
class Something(Model):
    name = fields.Text()
    notify = True


# for fields test
class Subscription(Model):
    key_name = fields.TreeIndex()


class User(Model):
    password = fields.Hash()


class Truck(Model):
    last_position = fields.Location()


# for form testing
class ShipForm(Form):
    name = fields.Text()
    code = fields.Text()


# for testing permissions
class Bunny(BoundedModel):
    name = fields.Text()

    @classmethod
    def prefix(cls):
        return 'bound'


# for testing validations
class MyForm(Form):
    field1 = fields.Text()
    field2 = fields.Text()

    @validation_rule
    def enforce_fields(data):
        if (data.field1 is None and data.field2 is None) or \
                (data.field1 is not None and data.field2 is not None):
            raise InvalidFieldError(field='field1')


class MyModel(Model):
    field1 = fields.Text(index=True, required=False, private=True)


# For testing relationships
class Pet(Model):
    name = fields.Text()
    owner = fields.ForeignIdRelation('coralillo.tests.models.Person', inverse='pets')


class Person(Model):
    name = fields.Text()
    pets = fields.SetRelation(Pet, on_delete='cascade', inverse='owner')


class UnattachedPerson(Model):
    name = fields.Text()
    pets = fields.SetRelation(Pet, on_delete='set_null', inverse='owner')


# For many to many relations
class Driver(Model):
    name = fields.Text()
    cars = fields.SetRelation('coralillo.tests.models.Car', inverse='drivers')


class Car(Model):
    name = fields.Text()
    drivers = fields.SetRelation(Driver, inverse='cars')


class Employee(Model):
    name = fields.Text()
    last_name = fields.Text()
    office = fields.ForeignIdRelation('coralillo.tests.models.Office', inverse='employees')


class Office(Model):
    name = fields.Text()
    address = fields.Text()
    employees = fields.SetRelation(Employee, inverse='office')


# for testing nested embedding
class A(Model):
    attr = fields.Text()
    bs = fields.SetRelation('coralillo.tests.models.B', inverse='a')


class B(Model):
    attr = fields.Text()
    a = fields.ForeignIdRelation(A, inverse='bs')
    cs = fields.SetRelation('coralillo.tests.models.C', inverse='b')


class C(Model):
    attr = fields.Text()
    b = fields.ForeignIdRelation(A, inverse='cs')


# For sorted set tests
class Admin(Model):
    name = fields.Text()
    logs = fields.SortedSetRelation('coralillo.tests.models.Log', sort_key='date', inverse='owner')


class Log(Model):
    data = fields.Text()
    date = fields.Datetime()
    owner = fields.ForeignIdRelation(Admin, inverse='logs')


def bound_models(eng):
    for name, cls in inspect.getmembers(sys.modules[__name__]):
        if inspect.isclass(cls):
            if issubclass(cls, Form):
                cls.set_engine(eng)



from .models import Office, Employee, A, B, C


def test_include_simplest(nrm):
    office = Office(
        name='Fleety',
        address='Springfield 132',
    ).save()

    assert office.to_json() == {
        'id': office.id,
        'name': office.name,
        'address': office.address,
        '_type': 'office',
    }

    assert office.to_json(include=['name']) == {
        'name': 'Fleety',
    }

    assert office.to_json(include=['id', 'name']) == {
        'id': office.id,
        'name': office.name,
    }

    assert office.to_json(include=['none']) == {}


def test_incude(nrm):
    office = Office(
        name='Fleety',
        address='Springfield 132',
    ).save()
    employee = Employee(
        name='Juan',
        last_name='Alduci',
    ).save()
    office.employees.set([employee])

    assert office.to_json(include=['id', 'employees']) == {
        'id': office.id,
        'employees': [employee.to_json()],
    }

    assert office.to_json(include=['id', 'employees.name']) == {
        'id': office.id,
        'employees': [employee.to_json(include=['name'])],
    }

    assert office.to_json(include=['id', 'employees.name', 'employees.id']) == {
        'id': office.id,
        'employees': [employee.to_json(include=['name', 'id'])],
    }


def test_include_foreignid_relation(nrm):
    office = Office(
        name='Fleety',
        address='Springfield 132',
    ).save()
    employee = Employee(
        name='Juan',
        last_name='Alduci',
    ).save()
    office.employees.set([employee])

    assert employee.to_json(include=['id', 'office']) == {
        'id': employee.id,
        'office': office.to_json(),
    }

    assert employee.to_json(include=['id', 'office.name']) == {
        'id': employee.id,
        'office': office.to_json(include=['name']),
    }

    assert employee.to_json(include=['id', 'office.name', 'office.id']) == {
        'id': employee.id,
        'office': office.to_json(include=['name', 'id']),
    }

    office.employees.remove(employee)

    assert employee.to_json(include=['id', 'office']) == {
        'id': employee.id,
        'office': None,
    }

    assert employee.to_json(include=['id', 'office.name']) == {
        'id': employee.id,
        'office': None,
    }

    assert employee.to_json(include=['id', 'office.name', 'office.id']) == {
        'id': employee.id,
        'office': None,
    }


def test_include_three_levels():
    a = A(attr='z').save()

    b = B(attr='z').save()
    b.a.set(a)

    c = C(attr='z').save()
    c.b.set(b)

    assert a.to_json(include=['attr', 'bs.cs']) == {
        'attr': 'z',
        'bs': [{
            'cs': [{
                '_type': 'c',
                'id': c.id,
                'attr': c.attr,
            }],
        }],
    }

    assert a.to_json(include=['attr', 'bs.cs.id']) == {
        'attr': 'z',
        'bs': [{
            'cs': [{
                'id': c.id,
            }],
        }],
    }


def test_include_asterisc():
    a = A(attr='z').save()

    b = B(attr='z').save()
    b.a.set(a)

    c = C(attr='z').save()
    c.b.set(b)

    assert a.to_json(include=['*', 'bs.cs.attr']) == {
        '_type': 'a',
        'id': a.id,
        'attr': 'z',
        'bs': [{
            'cs': [{
                'attr': c.attr,
            }],
        }],
    }



import pytest

from .models import Bunny


def test_allow_key(nrm, user):
    user.allow('a')

    assert nrm.redis.exists('user:{}:allow'.format(user.id))


def test_add_permissions(nrm, user):
    user.allow('a:b:c')
    assert nrm.redis.sismember(user.allow_key(), 'a:b:c')


def test_add_ignores_when_has_parent(nrm, user):
    user.allow('a')
    user.allow('a:b')
    user.allow('a:b:c')

    assert nrm.redis.sismember(user.allow_key(), 'a')
    assert not nrm.redis.sismember(user.allow_key(), 'a:b')
    assert not nrm.redis.sismember(user.allow_key(), 'a:b:c')

    user.allow('foo:var')
    user.allow('foo:var:log')
    assert not nrm.redis.sismember(user.allow_key(), 'foo')
    assert nrm.redis.sismember(user.allow_key(), 'foo:var')
    assert not nrm.redis.sismember(user.allow_key(), 'foo:var:log')


def test_add_deletes_lower(nrm, user):
    user.allow('a:b/v')
    user.allow('a/v')

    assert user.get_perms() == set(['a/v'])


def test_revoke_permission(nrm, user):
    user.allow('a:b')
    user.revoke('a:b')

    assert not nrm.redis.sismember(user.allow_key(), 'a:b')


def test_check_permission_inheritance(nrm, user):
    user.allow('a:b')

    assert user.is_allowed('a:b')
    assert user.is_allowed('a:b:c')

    assert not user.is_allowed('a')
    assert not user.is_allowed('a:d')


def test_can_carry_restrict(nrm, user):
    user.allow('org:fleet/view')

    assert user.is_allowed('org:fleet:somefleet/view')


def test_permission_key(nrm, user):
    assert user.permission() == 'user:{}'.format(user.id)
    assert user.permission(restrict='view') == 'user:{}/view'.format(user.id)


def test_minor_ignored_if_mayor(nrm, user):
    user.allow('org:fleet/view')
    user.allow('org:fleet:325234/view')

    assert user.get_perms() == set(['org:fleet/view'])


def test_perm_framework_needs_strings(nrm, user):
    class User:
        pass

    with pytest.raises(AssertionError):
        user.allow(User)

    with pytest.raises(AssertionError):
        user.is_allowed(User)

    with pytest.raises(AssertionError):
        user.revoke(User)


def test_permission_function(nrm, user):
    assert user.permission() == 'user:{}'.format(user.id)
    assert user.permission('eat') == 'user:{}/eat'.format(user.id)

    bunny = Bunny(name='doggo').save()
    assert bunny.permission() == 'bound:bunny:{}'.format(bunny.id)
    assert bunny.permission('walk') == 'bound:bunny:{}/walk'.format(bunny.id)


def test_delete_user_deletes_permission_bag(nrm, user):
    user.allow('foo')

    assert nrm.redis.exists('user:{}:allow'.format(user.id))

    user.delete()
    assert not nrm.redis.exists('user:{}:allow'.format(user.id))



from math import radians, sin, cos, asin, sqrt

EPSILON = 0.00001


def debyte_string(byte_string):
    if type(byte_string) == bytes:
        return byte_string.decode('utf8')

    return byte_string


def debyte_iterator(iterator):
    return map(debyte_string, iterator)


def debyte_set(iterator):
    return set(debyte_iterator(iterator))


def debyte_list(iterator):
    return list(debyte_iterator(iterator))


def debyte_tuple(iterator):
    return tuple(debyte_iterator(iterator))


def debyte_float(byte_float):
    return float(byte_float)


def debyte_hash(dictlike):
    return dict(map(
        debyte_tuple,
        dictlike.items()
    ))


class Location:

    def __init__(self, lon, lat):
        self.lon = lon
        self.lat = lat

    def distance(self, loc):
        """
        Calculate the great circle distance between two points
        on the earth (specified in decimal degrees)
        """
        assert type(loc) == type(self)

        # convert decimal degrees to radians
        lon1, lat1, lon2, lat2 = map(radians, [
            self.lon,
            self.lat,
            loc.lon,
            loc.lat,
        ])

        # haversine formula
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        a = sin(dlat / 2)**2 + cos(lat1) * cos(lat2) * sin(dlon / 2)**2
        c = 2 * asin(sqrt(a))
        r = 6371000  # Radius of earth in meters.
        return c * r

    def to_json(self):
        return {
            'lat': self.lat,
            'lon': self.lon,
        }

    def __eq__(self, other):
        if type(other) != Location:
            return False

        return abs(self.lat - other.lat) < EPSILON and abs(self.lon - other.lon) < EPSILON

    def __str__(self):
        return '<Location lat={} lon={}>'.format(self.lat, self.lon)



from coralillo.errors import ValidationErrors
import pytest

from .models import MyForm, MyModel


def test_validate_with_custom_rules(nrm):
    with pytest.raises(ValidationErrors):
        MyForm.validate(
            field1='po',
            field2='llo',
        )

    with pytest.raises(ValidationErrors):
        MyForm.validate()


def test_can_have_many_uniques_with_null_value(nrm):
    m1 = MyModel.validate().save()
    assert m1.field1 is None

    m2 = MyModel.validate().save()
    assert m2.field1 is None



from . import datamodel
from .datamodel import debyte_string
from .errors import MissingFieldError, InvalidFieldError, ReservedFieldError, NotUniqueFieldError, DeleteRestrictedError
from .hashing import make_password, is_hashed
from coralillo.queryset import QuerySet
from importlib import import_module
import datetime
import json
import re


class Field:
    ''' Defines a field of a model. Represents how to store this specific
    datatype in the redis database '''

    def __init__(self, *, name=None, index=False, required=True, default=None, private=False, regex=None, forbidden=None, allowed=None, fillable=True):
        # This field's value is mapped to the ID in a redis hash so you can Model.get_by(field, value)
        self.index = index

        # This field is required in validation
        self.required = required

        # This field's default value
        self.default = default

        # This field's value is not published in the JSON representation of the object
        self.private = private

        # A regular expresion that validates this field's value
        self.regex = regex

        # A set of forbidden values for this field
        self.forbidden = forbidden

        # The set of only allowed valies for this field
        self.allowed = allowed

        # This field can't be set via http
        self.fillable = fillable

        self.name = name

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if isinstance(self, Relation):
            return self.manager(instance)

        return instance.__dict__[self.name]

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value

    def value_or_default(self, value):
        ''' Returns the given value or the specified default value for this
        field '''
        if value is None:
            if callable(self.default):
                return self.default()
            else:
                return self.default

        return value

    def validate_required(self, value):
        ''' Validates the given value agains this field's 'required' property
        '''
        if self.required and (value is None or value == ''):
            raise MissingFieldError(self.name)

    def init(self, value):
        ''' Returns the value that will be set in the model when it is passed
        as an __init__ attribute '''
        return self.value_or_default(value)

    def recover(self, instance, data, redis):
        ''' Retrieve this field's value from the database '''
        value = data.get(self.name)

        if value is None or value == 'None':
            return None

        return str(value)

    def prepare(self, value):
        ''' Prepare this field's value to insert in database '''
        if value is None:
            return None

        return str(value)

    def to_json(self, value):
        ''' Format the value to be presented in json format '''
        return value

    def save(self, instance, value, redis):
        ''' Sets this field's value in the databse '''
        value = self.prepare(value)

        if value is not None:
            redis.hset(instance.key(), self.name, value)
        else:
            redis.hdel(instance.key(), self.name)

        if self.index:
            key = self.key(instance)

            if self.name in instance._old:
                redis.hdel(key, instance._old[self.name])

            if value is not None:
                redis.hset(key, value, instance.id)

    def _delete(self, instance, redis):
        ''' Deletes this field's value from the databse. Should be implemented
        in special cases '''
        if self.index:
            redis.hdel(self.key(instance), getattr(instance, self.name))

    def validate(self, instance, value, redis):
        '''
        Validates data obtained from a request and returns it in the apropiate
        format
        '''
        # cleanup
        if type(value) == str:
            value = value.strip()

        value = self.value_or_default(value)

        # validation
        self.validate_required(value)

        if self.regex and not re.match(self.regex, value, flags=re.ASCII):
            raise InvalidFieldError(self.name)

        if self.forbidden and value in self.forbidden:
            raise ReservedFieldError(self.name)

        if self.allowed and value not in self.allowed:
            raise InvalidFieldError(self.name)

        if self.index:
            key = self.key(instance)

            old = debyte_string(redis.hget(key, value)) if value is not None else None
            old_value = getattr(instance, self.name)

            if old is not None and old != instance.id:
                raise NotUniqueFieldError(self.name)
            elif old_value != value:
                instance._old[self.name] = old_value

        return value

    def key(self, obj):
        return obj.cls_key() + ':index_' + self.name


class Text(Field):
    pass


class TreeIndex(Field):

    def save(self, instance, value, redis):
        ''' Sets this fields value in the databse '''
        value = self.prepare(value)

        if value is not None:
            redis.hset(instance.key(), self.name, value)
        else:
            redis.hdel(instance.key(), self.name)

        key = self.key(instance)

        if self.name in instance._old:
            redis.hdel(key, instance._old[self.name])

        redis.sadd(key + ':' + value, instance.id)

    def _delete(self, instance, redis):
        ''' Deletes this field's value from the databse. Should be implemented
        in special cases '''
        value = getattr(instance, self.name)
        redis.srem(self.key(instance) + ':' + value, instance.id)

    def key(self, obj):
        return obj.cls_key() + ':tree_' + self.name


class Hash(Text):
    ''' A value that should be stored as a hash, for example a password '''

    def init(self, value):
        ''' hash passwords given in the constructor '''
        value = self.value_or_default(value)

        if value is None:
            return None

        if is_hashed(value):
            return value

        return make_password(value)

    def prepare(self, value):
        ''' Prepare this field's value to insert in database '''
        if value is None:
            return None

        if is_hashed(value):
            return value

        return make_password(value)

    def validate(self, instance, value, redis):
        ''' hash passwords given via http '''
        value = super().validate(instance, value, redis)

        if is_hashed(value):
            return value

        return make_password(value)


class Bool(Field):
    ''' A boolean value '''

    def validate(self, instance, value, redis):
        value = self.value_or_default(value)

        if value is None:
            return None

        if type(value) == bool:
            return value

        return value == 'true' or value == '1'

    def prepare(self, value):
        return str(value)

    def recover(self, instance, data, redis):
        value = data.get(self.name)

        if value is None or value == 'None':
            return None

        return value in ['True', 'true', '1', 1]


class Integer(Field):
    ''' An integer value '''

    def validate(self, instance, value, redis):
        value = self.value_or_default(value)

        self.validate_required(value)

        if value != 0 and not value:
            return None

        try:
            return int(value)
        except ValueError:
            raise InvalidFieldError(self.name)

    def recover(self, instance, data, redis):
        value = data.get(self.name)

        if value == '' or value is None or value == 'None':
            return None

        return int(value)

    def prepare(self, value):
        return str(value)


class Float(Field):

    def validate(self, instance, value, redis):
        value = self.value_or_default(value)

        self.validate_required(value)

        try:
            return float(value)
        except ValueError:
            raise InvalidFieldError(self.name)

    def recover(self, instance, data, redis):
        value = data.get(self.name)

        if value == '' or value is None or value == 'None':
            return None

        return float(value)

    def prepare(self, value):
        return str(value)


class Datetime(Field):
    ''' A datetime that can be used transparently as such in the model '''

    def validate(self, instance, value, redis):
        '''
        Validates data obtained from a request in ISO 8061 and returns it in Datetime data type
        '''

        value = self.value_or_default(value)

        self.validate_required(value)

        if value is None:
            return None

        if type(value) == str:
            try:
                value = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')
            except ValueError:
                raise InvalidFieldError(self.name)

        return value

    def recover(self, instance, data, redis):
        value = data.get(self.name)

        if not value or value == 'None':
            return None

        return datetime.datetime.utcfromtimestamp(int(value))

    def prepare(self, value):
        if value is None:
            return None

        return str(int(value.timestamp()))

    def to_json(self, value):
        if value is None:
            return None

        return value.replace(microsecond=0).isoformat() + 'Z'


class Location(Field):
    ''' A geolocation '''

    def prepare(self, value):
        return value

    def save(self, instance, value, redis):
        key = self.key(instance)

        if value is not None:
            redis.geoadd(key, value.lon, value.lat, instance.id)
        else:
            redis.zrem(key, instance.id)

    def _delete(self, instance, redis):
        key = self.key(instance)

        redis.zrem(key, instance.id)

    def to_json(self, value):
        if value is None:
            return None

        return value.to_json()

    def recover(self, instance, data, redis):
        key = self.key(instance)
        value = redis.geopos(key, instance.id)

        if not value:
            return None

        if value[0] is None:
            return None

        return datamodel.Location(*value[0])

    def validate(self, instance, value, redis):
        value = self.value_or_default(value)

        self.validate_required(value)

        if value is None:
            return None

        try:
            lon, lat = map(float, value.split(','))

            assert -180 < lon < 180
            assert -90 < lat < 90

            return datamodel.Location(lon, lat)
        except ValueError:
            raise InvalidFieldError(self.name)

    def key(self, obj):
        return obj.cls_key() + ':geo_' + self.name


class Dict(Field):
    ''' A dict that can be used transparently as such in the model '''

    def validate(self, instance, value, redis):
        if not value:
            return dict()

        try:
            value = json.loads(value)
        except json.decoder.JSONDecodeError:
            raise InvalidFieldError(self.name)

        return value

    def prepare(self, value):
        return value

    def save(self, instance, value, redis):
        key = self.key(instance)

        if value is not None:
            redis.delete(key)
            redis.hset(key, self.name, json.dumps(value))
        else:
            redis.delete(key)

    def _delete(self, instance, redis):
        key = self.key(instance)

        redis.delete(key)

    def to_json(self, value):
        if value is None:
            return dict()

        return value

    def recover(self, instance, data, redis):
        key = self.key(instance)

        try:
            value = json.loads(redis.hget(key, self.name))
        except TypeError:
            value = dict()

        return value

    def key(self, obj):
        return '{}:{}:dict_{}'.format(obj.cls_key(), obj.id, self.name)


def model_from_spec(modelspec):
    if type(modelspec) == str:
        pieces = modelspec.split('.')

        return getattr(import_module('.'.join(pieces[:-1])), pieces[-1])

    return modelspec


class Relation(Field):

    def __init__(self, model, *, private=False, on_delete='set_null', inverse=None):
        self.index = False
        self.modelspec = model
        self.private = private
        self.on_delete = on_delete
        self.inverse = inverse
        self.fillable = False

    def _delete(self, instance, redis):
        raise NotImplementedError()


class SingleRelation(Relation):
    pass


class ForeignIdRelation(SingleRelation):

    def __init__(self, model, *, private=False, on_delete='set_null', inverse=None):
        super().__init__(model, private=private, on_delete=on_delete, inverse=inverse)
        self.default = None

    def validate(self, instance, value, redis):
        if value is None:
            return None

        related_obj = model_from_spec(self.modelspec).get(value)

        if related_obj is None:
            raise InvalidFieldError(self.name)

        return related_obj

    def save(self, instance, value, redis):
        if value is not None:
            assert type(value) == model_from_spec(self.modelspec)
            redis.hset(instance.key(), self.name, value.id)

    def _delete(self, instance, redis):
        item = getattr(instance, self.name).get()

        if item is None:
            return

        if self.on_delete == 'restrict':
            raise DeleteRestrictedError('attempt to delete with relations and restrict flag')

        if self.on_delete == 'cascade':
            item.delete()
        elif self.inverse:
            getattr(item, self.inverse)._unrelate(instance, redis)

    def manager(self, instance):
        return SingleRelationManager(instance, self.inverse, self.modelspec, self.name)


class SingleRelationManager:

    def __init__(self, instance, inverse, modelspec, name):
        self.inverse = inverse
        self.instance = instance
        self.modelspec = modelspec
        self.name = name

    def _relate(self, obj, pipeline):
        pipeline.hset(self.instance.key(), self.name, obj.id)

    def _unrelate(self, obj, redis):
        redis.hdel(self.instance.key(), self.name, obj.id)

    def get(self):
        redis = self.instance.get_redis()
        value = debyte_string(redis.hget(self.instance.key(), self.name))

        return model_from_spec(self.modelspec).get(value)

    def set(self, obj):
        redis = self.instance.get_redis()
        prev = getattr(self.instance, self.name).get()

        if prev is not None and self.inverse:
            getattr(prev, self.inverse)._unrelate(self.instance, redis)

        if obj is None:
            redis.hdel(self.instance.key(), self.name)
            setattr(self.instance, self.name, None)
            return

        redis.hset(self.instance.key(), self.name, obj.id)

        if self.inverse:
            getattr(obj, self.inverse)._relate(self.instance, redis)


class MultipleRelationManager:

    def __init__(self, instance, relation_key, inverse, modelspec):
        self.inverse = inverse
        self.instance = instance
        self.relation_key = relation_key
        self.modelspec = modelspec

    def set(self, value):
        pipe = self.instance.get_redis().pipeline()

        pipe.delete(self.relation_key)

        self._relate_all(value, pipe)

        for related in value:
            if self.inverse:
                getattr(related, self.inverse)._relate(self.instance, pipe)

        pipe.execute()

    def add(self, obj):
        assert isinstance(obj, model_from_spec(self.modelspec))
        pipe = self.instance.get_redis().pipeline()

        self._relate(obj, pipe)

        if self.inverse:
            getattr(obj, self.inverse)._relate(self.instance, pipe)

        pipe.execute()

    def all(self, **kwargs):
        ''' Returns this relation '''
        redis = self.instance.get_redis()
        related = list(map(
            lambda id: model_from_spec(self.modelspec).get(debyte_string(id)),
            self.get_related_ids(redis, **kwargs)
        ))

        return related

    def remove(self, value):
        assert isinstance(value, model_from_spec(self.modelspec))
        redis = self.instance.get_redis()

        self._unrelate(value, redis)

        if self.inverse:
            getattr(value, self.inverse)._unrelate(self.instance, redis)

    def count(self):
        raise NotImplementedError('count is not implemented yet for this subclass of MultipleRelation')

    def q(self):
        raise NotImplementedError('q is not implemented for this subclass os MultipleRelation')

    def create(self, **kwargs):
        ''' Creates a new instance of the related model and relates it to the
        current model through this field '''
        raise NotImplementedError()

    def _relate_all(self, pipeline):
        raise NotImplementedError()

    def clear(self):
        ''' Clears all the relations of this field to another model '''
        for related in self.all():
            self.remove(related)


class SetRelationManager(MultipleRelationManager):

    def _relate(self, obj, redis):
        redis.sadd(self.relation_key, obj.id)

    def _relate_all(self, value, redis):
        redis.sadd(self.relation_key, *[r.id for r in value])

    def _unrelate(self, obj, redis):
        redis.srem(self.relation_key, obj.id)

    def get_related_ids(self, redis):
        return redis.smembers(self.relation_key)

    def count(self):
        return self.instance.get_redis().scard(self.relation_key)

    def q(self):
        redis = self.instance.get_redis()

        return QuerySet(model_from_spec(self.modelspec), redis.sscan_iter(self.relation_key))

    def __contains__(self, item):
        if not isinstance(item, model_from_spec(self.modelspec)):
            return False

        return self.instance.get_redis().sismember(self.relation_key, item.id)


class SortedSetRelationManager(MultipleRelationManager):

    def __init__(self, instance, relation_key, inverse, modelspec, sort_key):
        super().__init__(instance, relation_key, inverse, modelspec)
        self.sort_key = sort_key

    def _relate(self, obj, redis):
        field = getattr(model_from_spec(self.modelspec), self.sort_key)  # the field in the foreign model

        redis.zadd(self.relation_key, {
            obj.id: int(field.prepare(getattr(obj, self.sort_key))),
        })

    def _relate_all(self, value, redis):
        field = getattr(model_from_spec(self.modelspec), self.sort_key)  # the field in the foreign model

        mapping = {
            r.id: int(field.prepare(getattr(r, self.sort_key)))
            for r in value
        }

        redis.zadd(self.relation_key, mapping)

    def _unrelate(self, obj, redis):
        redis.zrem(self.relation_key, obj.id)

    def get_related_ids(self, redis, *, score=None):
        if score:
            return redis.zrangebyscore(self.relation_key, *score)

        return redis.zrange(self.relation_key, 0, -1)

    def count(self):
        redis = self.instance.get_redis()

        return redis.zcard(self.relation_key)

    def __contains__(self, item):
        if not isinstance(item, model_from_spec(self.modelspec)):
            return False

        redis = self.instance.get_redis()

        return redis.zscore(self.relation_key, item.id) is not None


class MultipleRelation(Relation):
    ''' Indicates that this field can associate with multiple objects of some other class '''

    def manager(self):
        raise NotImplementedError('Must be implemented in subclass')

    def _delete(self, instance, redis):
        key = self.key(instance)

        items = getattr(instance, self.name).all()

        if self.on_delete == 'restrict' and len(items) > 0:
            raise DeleteRestrictedError('attempt to delete with relations and restrict flag')

        for item in items:
            if self.on_delete == 'cascade':
                item.delete()
            elif self.on_delete == 'set_null' and self.inverse:
                getattr(item, self.inverse)._unrelate(instance, redis)

        redis.delete(key)


class SetRelation(MultipleRelation):
    ''' A relationship with another model where order doesn't matter '''

    def key(self, instance):
        return '{}:{}:srel_{}'.format(instance.cls_key(), instance.id, self.name)

    def manager(self, instance):
        return SetRelationManager(instance, self.key(instance), self.inverse, self.modelspec)


class SortedSetRelation(MultipleRelation):
    ''' A relationship with another model that ensures the same ordering every
    time '''

    def __init__(self, model, sort_key, **kwargs):
        super().__init__(model, **kwargs)
        self.sort_key = sort_key

    def key(self, instance):
        return '{}:{}:zrel_{}'.format(instance.cls_key(), instance.id, self.name)

    def manager(self, instance):
        return SortedSetRelationManager(instance, self.key(instance), self.inverse, self.modelspec, self.sort_key)



import os

import pytest

from coralillo import Engine, Model, fields
from coralillo.auth import PermissionHolder

from .models import bound_models


@pytest.fixture
def nrm():
    nrm = Engine(host=os.getenv('REDIS_HOST', 'localhost'))

    nrm.lua.drop(args=['*'])

    bound_models(nrm)

    return nrm


@pytest.fixture
def user(nrm):
    class User(Model, PermissionHolder):
        name = fields.Text()

        class Meta:
            engine = nrm

    return User(
        name='juan',
    ).save()
